from __future__ import annotations

import datetime
import enum
import functools
import logging
import os
import pathlib
import threading
import time
from pprint import pprint
from typing import Optional

import yaml
import zocalo.util.symlink
import zocalo.wrapper

import relion
from relion.cryolo_relion_it import cryolo_relion_it, dls_options, icebreaker_histogram
from relion.cryolo_relion_it.cryolo_relion_it import RelionItOptions
from relion.dbmodel.modeltables import (
    CryoemInitialModelTable,
    CTFTable,
    MotionCorrectionTable,
    ParticleClassificationGroupTable,
    ParticleClassificationTable,
    ParticlePickerTable,
    RelativeIceThicknessTable,
)

try:
    from relion.pipeline import PipelineRunner
except ImportError:
    PipelineRunner = None

logger = logging.getLogger("relion.zocalo.wrapper")

RelionStatus = enum.Enum("RelionStatus", "RUNNING SUCCESS FAILURE")

# to test:
# - create an empty directory
# - run following command to generate a recipe-wrapper:
#   dlstbx.find_in_ispyb -p 6844019 -f /dls_sw/apps/zocalo/live/recipes/ispyb-relion.json --recipe-pointer=2 --out=rw && replace "/dls/m12/data/2021/cm28212-1/processed" "$(pwd)/processed" "/dls/m12/data/2021/cm28212-1/tmp" "$(pwd)/tmp" '"$ispyb_autoprocprogram_id"' "83862530" < rw > rw-local
# - run the wrapper:
#   dlstbx.wrap --recipewrapper=rw-local --wrap=relion --offline -v


class RelionWrapper(zocalo.wrapper.BaseWrapper):
    def run(self):
        # Enable log messages for relion.*
        logging.getLogger("relion").setLevel(logging.INFO)

        # declare success here so we can mark a failure detected within processing thread
        # assume good things to start with
        success = True

        # Report python-relion package version
        self.status_thread.set_static_status_field("python-relion", relion.__version__)

        assert hasattr(self, "recwrap"), "No recipewrapper object found"

        self.params = self.recwrap.recipe_step["job_parameters"]
        self.working_directory = pathlib.Path(self.params["working_directory"])
        self.results_directory = pathlib.Path(self.params["results_directory"])

        # Here we cheat. Ultimately we want to run the processing inside the
        # 'working' directory, and then copy relevant output into the 'results'
        # directory as we go along. In the first instance we will just run Relion
        # inside the 'results' directory, and ignore the 'working' directory
        # completely.
        self.working_directory = self.results_directory

        # Create working and results directory
        self.working_directory.mkdir(parents=True, exist_ok=True)
        self.results_directory.mkdir(parents=True, exist_ok=True)
        if self.params.get("create_symlink"):
            # Create symbolic link above directories
            # We want users to go to the most recent execution, as they can stop
            # and restart processing from SynchWeb. Thus we overwrite symlinks.
            zocalo.util.symlink.create_parent_symlink(
                self.working_directory,
                self.params["create_symlink"],
                overwrite_symlink=True,
            )
            zocalo.util.symlink.create_parent_symlink(
                self.results_directory,
                self.params["create_symlink"],
                overwrite_symlink=True,
            )

        # Relion needs to have a 'Movies' link inside the working directory
        # pointing to the image files
        movielink = "Movies"
        os.symlink(self.params["image_directory"], self.working_directory / movielink)
        self.params["ispyb_parameters"]["import_images"] = os.path.join(
            movielink,
            pathlib.Path(self.params["ispyb_parameters"]["import_images"]).relative_to(
                self.params["image_directory"]
            ),
        )

        # Debug output
        pprint(self.params)

        # Select specific Cryolo version if so desired
        if self.params.get("cryolo_version"):
            os.environ["CRYOLO_VERSION"] = self.params["cryolo_version"]
            logger.info("Selected cryolo version %s", self.params["cryolo_version"])

        # Initialise number of imported files to 0
        imported_files = []
        mc_job_time_all_processed = None

        for k, v in self.params["ispyb_parameters"].items():
            if v.isnumeric():
                self.params["ispyb_parameters"][k] = int(v)
            elif v.lower() == "true":
                self.params["ispyb_parameters"][k] = True
            elif v.lower() == "false":
                self.params["ispyb_parameters"][k] = False
            else:
                try:
                    self.params["ispyb_parameters"][k] = float(v)
                except ValueError:
                    pass

        if self.params["ispyb_parameters"]["import_images"].endswith(".eer"):
            self.params["ispyb_parameters"]["motioncor_do_own"] = True
        if self.params["ispyb_parameters"].get("extract_small_boxsize"):
            self.params["ispyb_parameters"]["extract_downscale"] = True
        if self.params["ispyb_parameters"].get("extract2_small_boxsize"):
            self.params["ispyb_parameters"]["extract2_downscale"] = True
        pprint(self.params["ispyb_parameters"])

        movietype = self.params["ispyb_parameters"]["import_images"].split(".")[-1]

        self.opts = RelionItOptions()
        self.opts.update_from(vars(dls_options))
        self.opts.update_from(self.params["ispyb_parameters"])
        if os.getenv("RELION_CLUSTER_CONFIG"):
            with open(os.getenv("RELION_CLUSTER_CONFIG"), "r") as config:
                cluster_config = yaml.safe_load(config)
            self.opts.update_from(cluster_config)

        # Start Relion
        logger.info(f"Starting RELION version {self.params.get('relion_version', 3)}")
        self._relion_subthread = threading.Thread(
            target=self.start_relion,
            kwargs={
                "version": self.params.get("relion_version", 3),
                "movietype": movietype,
            },
            name="relion_subprocess_runner",
            daemon=True,
        )
        self._relion_subthread.start()

        relion_started = time.time()

        preprocess_check = self.results_directory / "RUNNING_PIPELINER_PREPROCESS"
        all_process_check = self.results_directory / "RUNNING_RELION_IT"

        relion_prj = relion.Project(
            self.working_directory,
            run_options=self.opts,
            message_constructors={
                "ispyb": construct_message,
                "images": images_msgs,
                "images_particles": images_particles_msgs,
            },
            version=self.params.get("relion_version", 3),
        )

        while not relion_prj.origin_present() or (
            not preprocess_check.is_file() and self.params.get("relion_version", 3) == 3
        ):
            time.sleep(0.5)
            if time.time() - relion_started > 10 * 60:
                break

        relion_prj.load()

        preproc_recently_run = False
        processing_ended = False
        should_send_icebreaker = True
        icebreaker_particles_star_file_found = False

        time_of_last_ib_hist_warning = 0

        while (
            self._relion_subthread.is_alive() or preprocess_check.is_file()
        ) and False not in [
            n.environment["status"]
            for n in relion_prj
            if n._out and n.name not in ("Class2D", "Class3D")
        ]:
            time.sleep(1)

            # logger.info("Looking for results")

            ispyb_command_list = []
            images_command_list = []
            images_particles_command_list = []

            if pathlib.Path(self.params["stop_file"]).is_file():
                logger.info("Stop file encountered")
                relion_prj.load()
                for job_path in relion_prj._job_nodes:
                    (
                        self.results_directory
                        / job_path.name
                        / "RELION_JOB_EXIT_ABORTED"
                    ).touch()
                for p in self.results_directory.glob("RUNNING_*"):
                    p.unlink()
                logger.info("Instructed Relion to stop. Terminating main loop.")
                break

            relion_prj.load()

            # Should only return results that have not previously been sent

            # for fr in relion_prj.results.fresh:
            #    curr_res = ispyb_results(fr.stage_object, fr.job_name, self.opts)
            #    ispyb_command_list.extend(curr_res)
            #    images_command_list.extend(images_msgs(fr.stage_object, fr.job_name))
            #    if curr_res:
            #        logger.info(f"Fresh results found for {fr.job_name}")

            # if ispyb_command_list:
            #    logger.info(
            #        "Sending commands like this: %s", str(ispyb_command_list[0])
            #    )
            #    self.recwrap.send_to(
            #        "ispyb", {"ispyb_command_list": ispyb_command_list}
            #    )
            #    logger.info("Sent %d commands to ISPyB", len(ispyb_command_list))

            # Should only return results that have not previously been sent

            for job_msg in relion_prj.messages:
                if job_msg.get("ispyb") and job_msg["ispyb"]:
                    logger.info(
                        f"Found results that look like this: {job_msg['ispyb'][0]}"
                    )
                    ispyb_command_list.extend(job_msg["ispyb"])
                if job_msg.get("images") and job_msg["images"]:
                    images_command_list.extend(job_msg["images"])
                if job_msg.get("images_particles") and job_msg["images_particles"]:
                    images_particles_command_list.extend(job_msg["images_particles"])

            if ispyb_command_list:
                # split up multi-part ISPyB messages into chunks of 200 to reduce load on the message broker
                multipart_limit = 200
                num_msgs = len(ispyb_command_list) // multipart_limit
                if len(ispyb_command_list) % multipart_limit:
                    num_msgs += 1
                for imsg in range(num_msgs):
                    if imsg == num_msgs - 1:
                        this_msg = ispyb_command_list[multipart_limit * imsg :]
                    else:
                        this_msg = ispyb_command_list[
                            multipart_limit * imsg : multipart_limit * (imsg + 1)
                        ]
                    logger.info("Sending commands like this: %s", str(this_msg[0]))
                    self.recwrap.send_to("ispyb", {"ispyb_command_list": this_msg})
                    logger.info("Sent %d commands to ISPyB", len(this_msg))

            for imgcmd in images_command_list:
                if imgcmd:
                    self.recwrap.send_to("images", imgcmd)

            for imgcmd in images_particles_command_list:
                if imgcmd:
                    imgcmd.update(
                        {
                            "angpix": self.opts.angpix,
                            "diameter": self.opts.mask_diameter,
                        }
                    )
                    self.recwrap.send_to("images_particles", imgcmd)

            ### Extract and send Icebreaker results as histograms if the Icebreaker grouping job has run
            if not self.opts.stop_after_ctf_estimation and (
                self.opts.do_class2d or self.opts.do_class3d
            ):
                attachment_list = []
                try:
                    pdf_file_path = icebreaker_histogram.create_pdf_histogram(
                        self.working_directory,
                        version=self.params.get("relion_version", 3.1),
                    )
                    json_file_path = icebreaker_histogram.create_json_histogram(
                        self.working_directory,
                        version=self.params.get("relion_version", 3.1),
                    )
                    if should_send_icebreaker and pdf_file_path and json_file_path:
                        attachment_list.append(
                            ispyb_attachment(json_file_path, "Graph")
                        )
                        attachment_list.append(ispyb_attachment(pdf_file_path, "Graph"))
                        logger.info(f"Sending ISPyB attachments {attachment_list}")
                        self.recwrap.send_to(
                            "ispyb", {"ispyb_command_list": attachment_list}
                        )
                        should_send_icebreaker = False
                        icebreaker_particles_star_file_found = True
                except (FileNotFoundError, OSError, RuntimeError, ValueError):
                    ib_hist_warning_time = time.time()
                    if ib_hist_warning_time - time_of_last_ib_hist_warning > 3600:
                        logger.debug(
                            "Error creating Icebreaker histogram.", exc_info=True
                        )
                        time_of_last_ib_hist_warning = ib_hist_warning_time

            # if Relion has been running too long stop loop of preprocessing jobs
            try:
                most_recent_movie = max(
                    p.stat().st_mtime
                    for p in pathlib.Path(self.params["image_directory"]).glob("**/*")
                )
            # if a file vanishes for some reason make sure that there is no crash and no exit
            except FileNotFoundError:
                most_recent_movie = time.time()

            # check if all imported files have been motion corrected
            # if they have then get the time stamp of the motion correction job
            # so that it can be checked all preprocessing jobs have run after it
            # only do this if the number of imported files has changed
            new_imported_files = relion_prj.get_imported()
            if new_imported_files != imported_files or not mc_job_time_all_processed:
                imported_files = new_imported_files
                mc_job_time_all_processed = self.check_processing_of_imports(
                    relion_prj, imported_files
                )

            if len(relion_prj._job_nodes) != 0 and len(relion_prj.preprocess) != 0:
                for job in relion_prj.preprocess:
                    job_end_time = job.environment["end_time_stamp"]
                    if job_end_time is None:
                        break
                    if (
                        datetime.datetime.timestamp(job_end_time) < most_recent_movie
                        or job.environment["job_count"] < 1
                    ):
                        break
                    # don't need to check if Import job has run recently for this bit
                    if mc_job_time_all_processed and "Import" not in job.name:
                        if (
                            datetime.datetime.timestamp(job_end_time)
                            < mc_job_time_all_processed
                        ):
                            preproc_recently_run = False
                            break
                else:
                    preproc_recently_run = True

            currtime = time.time()
            if (
                currtime - most_recent_movie
                > int(self.params["latest_movie_timeout"]) * 60
                and currtime - relion_started > 10 * 60
                and preprocess_check.is_file()
                and preproc_recently_run
                and mc_job_time_all_processed
            ):
                logger.info(
                    f"Ending preprocessing: current time {currtime}, most recent movie {most_recent_movie}, time at which all micrographs were processed {mc_job_time_all_processed}"
                )
                preprocess_check.unlink()

            processing_ended = self.check_whether_ended(relion_prj)
            if (
                currtime - most_recent_movie
                > int(self.params["latest_movie_timeout"]) * 60
                and currtime - relion_started > 10 * 60
                and all_process_check.is_file()
                and processing_ended
                and mc_job_time_all_processed
            ):
                logger.info(
                    f"Ending all processing: current time {currtime}, most recent movie {most_recent_movie}"
                )
                if all_process_check.is_file():
                    all_process_check.unlink()
                elif pathlib.Path(self.params["stop_file"]).is_file():
                    # User-initiated stop, still record it as failure in ISPyB,
                    # but don't write a warning to the log.
                    logger.info("Preprocessing stopped due to presence of stop file")
                    success = False
                else:
                    # if the running file is not there it was removed for a failure reason
                    logger.warning(
                        "Preprocessing exited unexpectedly. Relion wrapper returning failure status"
                    )
                    success = False

        if not icebreaker_particles_star_file_found:
            logger.warning("No particles.star file found for Icebreaker grouping.")
        logger.info("Done.")
        if success:
            success = False not in [
                n.environment["status"] for n in relion_prj if n._out
            ]

        if preprocess_check.is_file():
            preprocess_check.unlink()

        if all_process_check.is_file():
            all_process_check.unlink()

        return success

    @staticmethod
    def check_processing_of_imports(relion_prj, imported):
        try:
            checked_key = "job002"
            checks = [False for _ in range(len(imported))]
            for i, f in enumerate(imported):
                keys = [
                    (j.environment["job"], j)
                    for j in relion_prj._jobtype_nodes
                    if j.name == "MotionCorr"
                ]
                for key, job in keys:
                    if any(
                        f.split(".")[0] in p.micrograph_name.split(".")[0]
                        for p in job.environment["result"][key]
                    ):
                        checks[i] = True
                        checked_key = key
                        break
            if all(checks):
                completion_time = relion_prj._job_nodes.get_by_name(
                    "MotionCorr/" + checked_key
                ).environment["end_time_stamp"]
                if completion_time:
                    return datetime.datetime.timestamp(completion_time)
            return
        except (KeyError, AttributeError, RuntimeError, FileNotFoundError) as e:
            logger.debug(
                f"Exception encountered while checking whether imported files have been processed: {e}",
                exc_info=True,
            )
            return

    def start_relion(self, version: int = 3, movietype: str = "tiff"):
        print("Running RELION wrapper - stdout")
        logger.info("Running RELION wrapper - logger.info")

        # Write options to disk for a record of parameters used
        options_file = self.working_directory / cryolo_relion_it.OPTIONS_FILE
        logger.info(f"Writing all options to {options_file}")
        if os.path.isfile(options_file):
            logger.info(
                f"File {options_file} already exists; renaming old copy to {options_file}~"
            )
            os.rename(options_file, f"{options_file}~")
        with open(options_file, "w") as optfile:
            self.opts.print_options(optfile)

        success = False
        oldpwd = os.getcwd()
        try:
            version = int(version)
            os.chdir(self.working_directory)
            if version == 3:
                cryolo_relion_it.run_pipeline(self.opts)
            elif version == 4 and PipelineRunner:
                pipeline = PipelineRunner(
                    self.working_directory,
                    pathlib.Path(self.params["stop_file"]),
                    self.opts,
                    movietype=movietype,
                )
                pipeline.run(self.params["latest_movie_timeout"] * 60)
            else:
                logger.error(f"Unknown RELION version {version}: must be either 3 or 4")
                return False
            success = True
        except Exception as ex:
            logger.error(ex, exc_info=True)
        finally:
            os.chdir(oldpwd)

        logger.info("RELION thread done.")
        return success

    def check_whether_ended(self, proj):
        if len(proj._job_nodes) == 0 or None in [j.environment["status"] for j in proj]:
            return False
        check_time = time.time()
        return all(
            check_time - datetime.datetime.timestamp(j.environment["end_time_stamp"])
            > 10 * 60
            for j in proj
        )

    def create_synchweb_stop_file(self):
        pathlib.Path(self.params["stop_file"]).touch()

    def get_status(self, job_path):
        relion_stop_files = [
            "RELION_JOB_EXIT_SUCCESS",
            "RELION_EXIT_FAILURE",
            "RELION_JOB_ABORT_NOW",
            "RELION_EXIT_ABORTED",
        ]
        # synchweb_stop_files = [synchweb stop files list]
        # job_finished_files = [relion job finished files]
        for item in relion_stop_files:
            if (job_path / item).is_file():  # or synchweb_stop_file exists:
                return RelionStatus.SUCCESS
            else:
                return RelionStatus.RUNNING
            # if job_finished_file exists:


@functools.singledispatch
def images_msgs(table, primary_key, **kwargs):
    return []


@images_msgs.register(MotionCorrectionTable)
def _(table: MotionCorrectionTable, primary_key: int, **kwargs):
    return {
        "file": table.get_row_by_primary_key(primary_key)[
            "micrograph_snapshot_full_path"
        ].replace(".jpeg", ".mrc")
    }


@images_msgs.register(CTFTable)
def _(table: CTFTable, primary_key: int, **kwargs):
    return {
        "file": table.get_row_by_primary_key(primary_key)[
            "fft_theoretical_full_path"
        ].replace(".jpeg", ".ctf")
    }


@images_msgs.register(ParticleClassificationGroupTable)
def _(table: ParticleClassificationGroupTable, primary_key: int, **kwargs):
    image_path = table.get_row_by_primary_key(primary_key)["class_images_stack"]
    if not image_path:
        return {}
    return {"file": image_path, "all_frames": "true"}


@functools.singledispatch
def images_particles_msgs(table, primary_key, **kwargs):
    return []


@images_particles_msgs.register(ParticlePickerTable)
def _(table: ParticlePickerTable, primary_key: int, **kwargs):
    mc_image_path = table.get_row_by_primary_key(primary_key)[
        "mc_image_full_path"
    ].replace(".jpeg", ".mrc")
    parpick_image_path = table.get_row_by_primary_key(primary_key)[
        "summary_image_full_path"
    ]
    if not mc_image_path or not parpick_image_path:
        return {}
    coords = table.get_row_by_primary_key(primary_key)["particle_coordinates"]
    return {"file": mc_image_path, "outfile": parpick_image_path, "coordinates": coords}


@functools.singledispatch
def construct_message(table, primary_key, resend=False, unsent_appended=None):
    raise ValueError(f"{table!r} is not a known Table")


@construct_message.register(MotionCorrectionTable)
def _(
    table: MotionCorrectionTable,
    primary_key: int,
    resend: bool = False,
    unsent_appended: Optional[dict] = None,
):
    row = table.get_row_by_primary_key(primary_key)
    drift_data = row["drift_data"]
    buffered = ["motion_correction_id", "drift_data"]
    buffer_store = row["motion_correction_id"]
    drift_frames = [(frame.frame, frame.deltaX, frame.deltaY) for frame in drift_data]
    if resend:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {"motion_correction_id": buffer_store},
            "buffer_command": {
                "ispyb_command": "insert_motion_correction",
                **{k: v for k, v in row.items() if k not in buffered},
                "drift_frames": drift_frames,
            },
        }
    else:
        results = {
            "ispyb_command": "buffer",
            "buffer_command": {
                "ispyb_command": "insert_motion_correction",
                **{k: v for k, v in row.items() if k not in buffered},
                "drift_frames": drift_frames,
            },
            "buffer_store": buffer_store,
        }
    return results


@construct_message.register(CTFTable)
def _(
    table: CTFTable,
    primary_key: int,
    resend: bool = False,
    unsent_appended: Optional[dict] = None,
):
    row = table.get_row_by_primary_key(primary_key)
    buffered = ["motion_correction_id", "ctf_id"]
    buffer_store = row["ctf_id"]
    buffer_lookup = row["motion_correction_id"]
    if resend:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {
                "motion_correction_id": buffer_lookup,
                "ctf_id": buffer_store,
            },
            "buffer_command": {
                "ispyb_command": "insert_ctf",
                **{k: v for k, v in row.items() if k not in buffered},
            },
        }
    else:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {
                "motion_correction_id": buffer_lookup,
            },
            "buffer_command": {
                "ispyb_command": "insert_ctf",
                **{k: v for k, v in row.items() if k not in buffered},
            },
            "buffer_store": buffer_store,
        }
    return results


@construct_message.register(ParticlePickerTable)
def _(
    table: ParticlePickerTable,
    primary_key: int,
    resend: bool = False,
    unsent_appended: Optional[dict] = None,
):
    row = table.get_row_by_primary_key(primary_key)
    buffered = [
        "first_motion_correction_id",
        "particle_picker_id",
        "particle_coordinates",
    ]
    buffer_store = row["particle_picker_id"]
    buffer_lookup = row["first_motion_correction_id"]
    if resend:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {
                "motion_correction_id": buffer_lookup,
                "particle_picker_id": buffer_store,
            },
            "buffer_command": {
                "ispyb_command": "insert_particle_picker",
                **{k: v for k, v in row.items() if k not in buffered},
            },
        }
    else:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {
                "motion_correction_id": buffer_lookup,
            },
            "buffer_command": {
                "ispyb_command": "insert_particle_picker",
                **{k: v for k, v in row.items() if k not in buffered},
            },
            "buffer_store": buffer_store,
        }
    return results


@construct_message.register(ParticleClassificationGroupTable)
def _(
    table: ParticleClassificationGroupTable,
    primary_key: int,
    resend: bool = False,
    unsent_appended: Optional[dict] = None,
):
    row = table.get_row_by_primary_key(primary_key)
    buffered = ["particle_picker_id", "particle_classification_group_id"]
    buffer_store = row["particle_classification_group_id"]
    buffer_lookup = row["particle_picker_id"]
    if resend:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {
                "particle_picker_id": buffer_lookup,
                "particle_classification_group_id": buffer_store,
            },
            "buffer_command": {
                "ispyb_command": "insert_particle_classification_group",
                **{k: v for k, v in row.items() if k not in buffered},
            },
        }
    else:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {
                "particle_picker_id": buffer_lookup,
            },
            "buffer_command": {
                "ispyb_command": "insert_particle_classification_group",
                **{k: v for k, v in row.items() if k not in buffered},
            },
            "buffer_store": buffer_store,
        }
    return results


@construct_message.register(ParticleClassificationTable)
def _(
    table: ParticleClassificationTable,
    primary_key: int,
    resend: bool = False,
    unsent_appended: Optional[dict] = None,
):
    row = table.get_row_by_primary_key(primary_key)
    buffered = ["particle_classification_group_id", "particle_classification_id"]
    buffer_store = row["particle_classification_id"]
    buffer_lookup = row["particle_classification_group_id"]
    if resend:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {
                "particle_classification_group_id": buffer_lookup,
                "particle_classification_id": buffer_store,
            },
            "buffer_command": {
                "ispyb_command": "insert_particle_classification",
                **{k: v for k, v in row.items() if k not in buffered},
            },
        }
    else:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {
                "particle_classification_group_id": buffer_lookup,
            },
            "buffer_command": {
                "ispyb_command": "insert_particle_classification",
                **{k: v for k, v in row.items() if k not in buffered},
            },
            "buffer_store": buffer_store,
        }
    return results


@construct_message.register(CryoemInitialModelTable)
def _(
    table: CryoemInitialModelTable,
    primary_key: int,
    resend: bool = False,
    unsent_appended: Optional[dict] = None,
):
    if unsent_appended is None:
        unsent_appended = {}
    row = table.get_row_by_primary_key(primary_key)
    class_ids = unsent_appended.get("particle_classification_id", [])
    # class_ids = row["particle_classification_id"]
    buffer_store = row["cryoem_initial_model_id"]
    if not isinstance(class_ids, list):
        class_ids = [class_ids]
    results = []
    if resend:
        for i, class_id in enumerate(class_ids):
            buffered = ["particle_classification_id", "cryoem_initial_model_id"]
            this_result = {
                "ispyb_command": "buffer",
                "buffer_lookup": {
                    "particle_classification_id": class_id,
                    "cryoem_initial_model_id": buffer_store,
                },
                "buffer_command": {
                    "ispyb_command": "insert_cryoem_initial_model",
                    **{k: v for k, v in row.items() if k not in buffered},
                },
            }
            results.append(this_result)
    else:
        for i, class_id in enumerate(class_ids):
            buffered = ["particle_classification_id", "cryoem_initial_model_id"]
            this_result = {
                "ispyb_command": "buffer",
                "buffer_lookup": {
                    "particle_classification_id": class_id,
                },
                "buffer_command": {
                    "ispyb_command": "insert_cryoem_initial_model",
                    **{k: v for k, v in row.items() if k not in buffered},
                },
            }
            if i == 0:
                this_result["buffer_store"] = buffer_store
            else:
                this_result["buffer_lookup"]["cryoem_initial_model_id"] = buffer_store
            results.append(this_result)
    return results


@construct_message.register(RelativeIceThicknessTable)
def _(
    table: RelativeIceThicknessTable,
    primary_key: int,
    resend: bool = False,
    unsent_appended: Optional[dict] = None,
):
    row = table.get_row_by_primary_key(primary_key)
    buffered = ["motion_correction_id", "relative_ice_thickness_id"]
    buffer_store = row["relative_ice_thickness_id"]
    buffer_lookup = row["motion_correction_id"]
    if resend:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {
                "motion_correction_id": buffer_lookup,
                "relative_ice_thickness_id": buffer_store,
            },
            "buffer_command": {
                "ispyb_command": "insert_relative_ice_thickness",
                **{k: v for k, v in row.items() if k not in buffered},
            },
        }
    else:
        results = {
            "ispyb_command": "buffer",
            "buffer_lookup": {
                "motion_correction_id": buffer_lookup,
            },
            "buffer_command": {
                "ispyb_command": "insert_relative_ice_thickness",
                **{k: v for k, v in row.items() if k not in buffered},
            },
            "buffer_store": buffer_store,
        }
    return results


def ispyb_attachment(attachment_path_object, file_type):
    return {
        "ispyb_command": "add_program_attachment",
        "file_name": os.fspath(attachment_path_object.name),
        "file_path": os.fspath(attachment_path_object.parent),
        "file_type": file_type,
    }



from __future__ import annotations

from sqlalchemy import TIMESTAMP, Column, ForeignKey, String
from sqlalchemy.dialects.mysql import INTEGER
from sqlalchemy.orm import declarative_base, relationship

Base = declarative_base()


class ZocaloBuffer(Base):
    __tablename__ = "ZocaloBuffer"

    AutoProcProgramID = Column(
        INTEGER(10),
        primary_key=True,
        comment="Reference to an existing AutoProcProgram",
        autoincrement=False,
        nullable=False,
    )
    UUID = Column(
        INTEGER(10),
        primary_key=True,
        comment="AutoProcProgram-specific unique identifier",
        autoincrement=False,
        nullable=False,
    )
    Reference = Column(
        INTEGER(10),
        comment="Context-dependent reference to primary key IDs in other ISPyB tables",
    )


class RelionPipelineInfo(Base):
    __tablename__ = "RelionPipelineInfo"

    pipeline_id = Column(INTEGER(10), primary_key=True)
    image_x = Column(
        INTEGER(5),
        comment="Number of pixels in the x direction",
        autoincrement=False,
    )
    image_y = Column(
        INTEGER(5),
        comment="Number of pixels in the y direction",
        autoincrement=False,
    )
    microscope = Column(
        String(10),
        nullable=False,
        comment="Microscope name",
    )
    project_path = Column(
        String(250),
        nullable=False,
        comment="Path to the project directory",
    )


class ClusterJobInfo(Base):
    __tablename__ = "ClusterJobInfo"

    cluster = Column(
        String(250),
        nullable=False,
        comment="Name of cluster",
        primary_key=True,
    )
    cluster_id = Column(
        INTEGER(10),
        primary_key=True,
        comment="ID of the cluster job",
        autoincrement=False,
        nullable=False,
    )
    auto_proc_program_id = Column(
        INTEGER(10),
        comment="Reference to the AutoProcProgram the cluster job is attached to",
        autoincrement=False,
    )
    start_time = Column(
        TIMESTAMP,
        comment="Start time of cluster job",
    )
    end_time = Column(
        TIMESTAMP,
        comment="End time of cluster job",
    )


class RelionJobInfo(Base):
    __tablename__ = "RelionJobInfo"

    job_id = Column(INTEGER(10), primary_key=True)
    cluster_id = Column(
        INTEGER(10),
        comment="ID of the cluster job",
        autoincrement=False,
    )
    relion_start_time = Column(
        TIMESTAMP,
        comment="Start time of Relion job",
    )
    num_micrographs = Column(
        INTEGER(10),
        comment="Number of micrographs processed by the job if applicable",
        autoincrement=False,
    )
    job_name = Column(
        String(250),
        nullable=False,
        comment="Name of Relion job",
    )
    pipeline_id = Column(ForeignKey("RelionPipelineInfo.pipeline_id"), index=True)
    RelionPipelineInfo = relationship("RelionPipelineInfo")


def buffer_url() -> str:
    import ispyb.sqlalchemy

    sqlalchemy_url = ispyb.sqlalchemy.url()
    local_url = "/".join(sqlalchemy_url.split("/")[:-1]) + "/zocalo"
    return local_url



from __future__ import annotations

import argparse
import os
import pathlib

from relion.cryolo_relion_it import cryolo_relion_it, dls_options
from relion.cryolo_relion_it.cryolo_relion_it import RelionItOptions

try:
    from relion.pipeline import PipelineRunner
except ModuleNotFoundError:
    PipelineRunner = None


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--dir",
        help="Working directory where Relion will run",
        dest="working_directory",
        default=".",
    )
    parser.add_argument(
        "-p",
        "--param",
        help="Set a key, value pair of options as key=value",
        action="append",
        dest="params",
    )
    parser.add_argument(
        "-f", "--file", help="Load options from file", dest="options_file"
    )
    parser.add_argument(
        "-m",
        "--movies",
        help="Directory to the directory to make Movies symlink to",
        dest="movies_dir",
    )
    parser.add_argument(
        "--version",
        help="Relion version; options are 3.1 or 4",
        dest="version",
        type=float,
        choices=[3.1, 4],
        default=3.1,
    )
    parser.add_argument(
        "--timeout",
        help="Stop importing movies if no new movies have been seen after this many seconds",
        dest="timeout",
        type=int,
        default=2 * 24 * 3600,
    )
    parser.add_argument(
        "--continue",
        help="Continue from previous pipeline",
        action="store_true",
        dest="continue_pipeline",
    )
    parser.add_argument(
        "--local",
        help="Turn off all cluster queue submission",
        dest="local",
        action="store_true",
    )
    args = parser.parse_args()

    opts = RelionItOptions()
    opts.update_from(vars(dls_options))

    if args.options_file:
        opts.update_from_file(args.options_file)

    if args.params:
        params_list = [p.split("=") for p in args.params]
        params_dict = {p[0]: p[1] for p in params_list}
        for k, v in params_dict.items():
            if v.isnumeric():
                params_dict[k] = int(v)
            elif v.lower() == "true":
                params_dict[k] = True
            elif v.lower() == "false":
                params_dict[k] = False
            else:
                try:
                    params_dict[k] = float(v)
                except ValueError:
                    pass
    else:
        params_dict = {}

    opts.update_from(params_dict)
    if not args.continue_pipeline:
        options_file = (
            pathlib.Path(args.working_directory) / cryolo_relion_it.OPTIONS_FILE
        )
        if os.path.isfile(options_file):
            os.rename(options_file, f"{options_file}~")
        with open(options_file, "w") as optfile:
            opts.print_options(optfile)

        if args.movies_dir is not None:
            (pathlib.Path(args.working_directory) / "Movies").symlink_to(
                args.movies_dir
            )

    abs_working = pathlib.Path(args.working_directory).resolve()
    os.chdir(args.working_directory)

    if args.version == 3.1:
        cryolo_relion_it.run_pipeline(opts)
    elif args.version == 4:
        if not PipelineRunner:
            exit("Relion 4 support requires the CCP-EM pipeliner package")
        suffix = ""
        for movie in (abs_working / "Movies").glob("**/*"):
            if movie.is_file():
                suffix = movie.suffix
                break
        if not suffix:
            raise ValueError(
                f"Movie suffix could not be determined from the files in {abs_working / 'Movies'}"
            )
        pipeline = PipelineRunner(
            abs_working,
            abs_working / "stop.stop",
            opts,
            movietype=suffix,
            restarted=args.continue_pipeline,
            local=args.local,
        )
        pipeline.run(args.timeout)



from __future__ import annotations

import csv
import logging
import pathlib
from collections import namedtuple

from relion._parser.jobtype import JobType

logger = logging.getLogger("relion._parser.relativeicethickness")

RelativeIceThicknessMicrograph = namedtuple(
    "RelativeIceThicknessMicrograph",
    ["micrograph_path", "minimum", "q1", "median", "q3", "maximum"],
)

RelativeIceThicknessMicrograph.__doc__ = "Relative ice thickness data for a micrograph."
RelativeIceThicknessMicrograph.minimum.__doc__ = "Minimum ice thickness. Unitless"
RelativeIceThicknessMicrograph.q1.__doc__ = "Quartile 1. Unitless"
RelativeIceThicknessMicrograph.median.__doc__ = "Median ice thickness. Unitless"
RelativeIceThicknessMicrograph.q3.__doc__ = "Quartile 3. Unitless"
RelativeIceThicknessMicrograph.maximum.__doc__ = "Maximum ice thickness. Unitless"
RelativeIceThicknessMicrograph.micrograph_path.__doc__ = "Micrograph path"


class RelativeIceThickness(JobType):
    def __eq__(self, other):
        if isinstance(other, RelativeIceThickness):
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.RelativeIceThickness", self._basepath))

    def __repr__(self):
        return f"RelativeIceThickness({repr(str(self._basepath))})"

    def __str__(self):
        return f"<RelativeIceThickness parser at {self._basepath}>"

    @property
    def job_number(self):
        jobs = [x.name for x in self._basepath.iterdir()]
        return jobs

    def _load_job_directory(self, jobdir):
        try:
            ice_dict = self.csv_to_dict(self._basepath / jobdir / "five_figs_test.csv")
        except (FileNotFoundError, OSError, RuntimeError, ValueError):
            return []
        list_micrograph_path = ice_dict["path"]
        list_minimum = ice_dict["min"]
        list_q1 = ice_dict["q1"]
        list_median = ice_dict["q2=median"]
        list_q3 = ice_dict["q3"]
        list_maximum = ice_dict["max"]

        micrograph_list = []
        for j in range(len(list_micrograph_path)):
            mic_path_parts = list(pathlib.Path(list_micrograph_path[j]).parts)
            ib_input_position = mic_path_parts.index("IB_input")
            micrograph_list.append(
                RelativeIceThicknessMicrograph(
                    str(
                        pathlib.Path(*mic_path_parts[(ib_input_position + 1) :])
                    ).replace("_grouped", "")
                    + ".mrc",  # convert the full path to the similar path for Motion Correction micrograph
                    list_minimum[j],
                    list_q1[j],
                    list_median[j],
                    list_q3[j],
                    list_maximum[j],
                )
            )
        return micrograph_list

    def csv_to_dict(self, file_path):
        with open(file_path, newline="") as csvfile:
            list_row_dicts = list(csv.DictReader(csvfile))
            combi_dict = {}
            for row_dict in list_row_dicts:
                for k, v in row_dict.items():
                    combi_dict.setdefault(k, []).append(v)
            return combi_dict

    @staticmethod
    def db_unpack(micrograph_list):
        res = [
            {
                "micrograph_full_path": micrograph.micrograph_path,
                "minimum": micrograph.minimum,
                "q1": micrograph.q1,
                "median": micrograph.median,
                "q3": micrograph.q3,
                "maximum": micrograph.maximum,
            }
            for micrograph in micrograph_list
        ]
        return res



from __future__ import annotations

from relion.node.graph import Graph


class DBGraph(Graph):
    def __init__(self, name, node_list, auto_connect=False):
        super().__init__(name, node_list, auto_connect=auto_connect)
        self.environment["source"] = self.name
        for n in self._node_list:
            for tab in n.tables:
                tab._last_update[self.name] = 0

    def __call__(self, *args, **kwargs):
        res = super().__call__(*args, **kwargs)
        collapsed_res = []
        if isinstance(res, dict):
            for curr_res in res.values():
                if curr_res is not None and curr_res not in collapsed_res:
                    collapsed_res.append(curr_res)
        elif isinstance(res, list):
            for el in res:
                if el is None:
                    continue
                if not el:
                    continue
                for curr_res in el.values():
                    if curr_res is not None and curr_res not in collapsed_res:
                        collapsed_res.append(curr_res)
        return collapsed_res

    def update_times(self, source=None):
        times = []
        for n in self._node_list:
            times.extend(n.update_times(source))
        return times

    def message(self, constructor=None):
        messages = []
        for node in self._node_list:

            messages.extend(node.message, constructor)
        return messages



from __future__ import annotations

import logging
import pathlib
from collections import namedtuple

from relion._parser.jobtype import JobType

logger = logging.getLogger("relion._parser.autopick")

ParticlePickerInfo = namedtuple(
    "ParticlePickerInfo",
    [
        "number_of_particles",
        "micrograph_full_path",
        "mc_image_full_path",
        "first_micrograph_name",
        "highlighted_micrograph",
        "coordinates",
        "job",
    ],
)

ParticleCacheRecord = namedtuple(
    "ParticleCacheRecord",
    [
        "data",
        "file_size",
    ],
)


class AutoPick(JobType):
    def __init__(self, path, particle_cache=None):
        super().__init__(path)
        self._particle_cache = particle_cache or {}

    def __eq__(self, other):
        if isinstance(other, AutoPick):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.AutoPick", self._basepath))

    def __repr__(self):
        return f"AutoPick({repr(str(self._basepath))})"

    def __str__(self):
        return f"<AutoPick parser at {self._basepath}>"

    def _load_job_directory(self, jobdir):
        try:
            file = self._read_star_file(jobdir, "summary.star")
        except (RuntimeError, FileNotFoundError, OSError, ValueError):
            return []

        info_table = self._find_table_from_column_name("_rlnGroupNrParticles", file)
        if info_table is None:
            logger.debug(f"_rlnGroupNrParticles not found in file {file}")
            return []

        all_particles = self.parse_star_file("_rlnGroupNrParticles", file, info_table)
        # num_particles = sum([int(n) for n in all_particles])

        mc_micrographs = self.parse_star_file("_rlnMicrographName", file, info_table)

        first_mc_micrograph = mc_micrographs[0]

        particle_picker_info = []
        for mic, np in zip(mc_micrographs, all_particles):
            coords = self._get_particle_info(jobdir, pathlib.Path(mic))
            mic_parts = pathlib.Path(mic).parts
            highlighted_micrograph = (
                self._basepath
                / jobdir
                / pathlib.Path(mic)
                .relative_to(pathlib.Path(mic_parts[0]) / mic_parts[1])
                .with_suffix(".jpeg")
            )
            particle_picker_info.append(
                ParticlePickerInfo(
                    int(np),
                    mic,
                    str(self._basepath.parent / mic).replace(".mrc", ".jpeg"),
                    first_mc_micrograph,
                    str(highlighted_micrograph),
                    coords,
                    jobdir,
                )
            )

        return particle_picker_info

    def _get_particle_info(self, jobdir, micrograph):
        particle_data = []
        mic_parts = micrograph.parts
        mc_job_path = pathlib.Path(mic_parts[0]) / mic_parts[1]
        particle_star_file = pathlib.Path(
            str(micrograph.relative_to(mc_job_path).with_suffix(".star")).replace(
                micrograph.stem, micrograph.stem + "_autopick"
            )
        )
        if self._particle_cache.get(jobdir):
            if self._particle_cache[jobdir].get(micrograph):
                try:
                    if (
                        self._particle_cache[jobdir][micrograph].file_size
                        == (self._basepath / jobdir / particle_star_file).stat().st_size
                    ):
                        return self._particle_cache[jobdir][micrograph].data
                except FileNotFoundError:
                    logger.debug(
                        "Could not find expected file containing particle data",
                        exc_info=True,
                    )
                    return []
        else:
            self._particle_cache[jobdir] = {}
        try:
            particle_star = self._read_star_file(jobdir, particle_star_file)
        except (FileNotFoundError, RuntimeError, ValueError):
            return particle_data
        try:
            info_table = self._find_table_from_column_name(
                "_rlnCoordinateX", particle_star
            )
        except (FileNotFoundError, RuntimeError, ValueError):
            return particle_data
        if info_table is None:
            logger.debug(
                f"_rlnMicrographFrameNumber not found in file {particle_star_file}"
            )
            return particle_data
        xs = self.parse_star_file("_rlnCoordinateX", particle_star, info_table)
        ys = self.parse_star_file("_rlnCoordinateY", particle_star, info_table)
        for x, y in zip(xs, ys):
            particle_data.append((x, y))
        try:
            self._particle_cache[jobdir][micrograph] = ParticleCacheRecord(
                particle_data,
                (self._basepath / jobdir / particle_star_file).stat().st_size,
            )
        except FileNotFoundError:
            return []
        return particle_data

    @staticmethod
    def for_cache(partpickinfo):
        return str(partpickinfo.number_of_particles)

    @staticmethod
    def db_unpack(partpickinfo):
        res = [
            {
                "number_of_particles": pi.number_of_particles,
                "job_string": pi.job,
                "micrograph_full_path": pi.micrograph_full_path,
                "mc_image_full_path": pi.mc_image_full_path,
                "summary_image_full_path": pi.highlighted_micrograph,
                "particle_coordinates": pi.coordinates,
            }
            for pi in partpickinfo
        ]
        return res



from __future__ import annotations

import logging
from collections import namedtuple

from relion._parser.jobtype import JobType

logger = logging.getLogger("relion._parser.ctffind")

CTFMicrograph = namedtuple(
    "CTFMicrograph",
    [
        "micrograph_name",
        "astigmatism",
        "defocus_u",
        "defocus_v",
        "defocus_angle",
        "max_resolution",
        "fig_of_merit",
        "amp_contrast",
        "diagnostic_plot_path",
    ],
)
CTFMicrograph.__doc__ = "Contrast Transfer Function stage."
CTFMicrograph.astigmatism.__doc__ = "Estimated astigmatism. Units angstrom (A)."
CTFMicrograph.micrograph_name.__doc__ = "Micrograph name. Useful for reference."
CTFMicrograph.defocus_u.__doc__ = (
    "Averaged with Defocus V to give estimated defocus. Units angstrom (A)."
)
CTFMicrograph.defocus_v.__doc__ = (
    "Averaged with Defocus U to give estimated defocus. Units angstrom (A)."
)
CTFMicrograph.defocus_angle.__doc__ = "Estimated angle of astigmatism."
CTFMicrograph.max_resolution.__doc__ = (
    "Maximum resolution that the software can detect. Units angstrom (A)."
)
CTFMicrograph.fig_of_merit.__doc__ = (
    "Figure of merit/CC/correlation value. Confidence of the defocus estimation."
)
CTFMicrograph.amp_contrast.__doc__ = "Amplitude contrast."
CTFMicrograph.diagnostic_plot_path.__doc__ = (
    "Path to the CTF diagnostic (fit/data comparison) plot (jpeg)."
)


class CTFFind(JobType):
    def __eq__(self, other):
        if isinstance(other, CTFFind):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.CTFFind", self._basepath))

    def __repr__(self):
        return f"CTFFind({repr(str(self._basepath))})"

    def __str__(self):
        return f"<CTFFind parser at {self._basepath}>"

    @property
    def job_number(self):
        jobs = [x.name for x in self._basepath.iterdir()]
        return jobs

    def _load_job_directory(self, jobdir):
        try:
            file = self._read_star_file(jobdir, "micrographs_ctf.star")
        except (FileNotFoundError, OSError, RuntimeError, ValueError):
            return []

        info_table = self._find_table_from_column_name("_rlnCtfAstigmatism", file)
        if info_table is None:
            return []

        astigmatism = self.parse_star_file("_rlnCtfAstigmatism", file, info_table)
        defocus_u = self.parse_star_file("_rlnDefocusU", file, info_table)
        defocus_v = self.parse_star_file("_rlnDefocusV", file, info_table)
        defocus_angle = self.parse_star_file("_rlnDefocusAngle", file, info_table)
        max_resolution = self.parse_star_file("_rlnCtfMaxResolution", file, info_table)
        fig_of_merit = self.parse_star_file("_rlnCtfFigureOfMerit", file, info_table)

        micrograph_name = self.parse_star_file("_rlnMicrographName", file, info_table)
        ctf_img_path = self.parse_star_file("_rlnCtfImage", file, info_table)

        info_table = self._find_table_from_column_name("_rlnAmplitudeContrast", file)

        amp_contrast = self.parse_star_file("_rlnAmplitudeContrast", file, info_table)

        micrograph_list = []
        for j in range(len(micrograph_name)):
            plot_path = (
                str(self._basepath.parent / ctf_img_path[j])
                .split(":")[0]
                .replace(".ctf", ".jpeg")
            )
            micrograph_list.append(
                CTFMicrograph(
                    micrograph_name[j],
                    astigmatism[j],
                    defocus_u[j],
                    defocus_v[j],
                    defocus_angle[j],
                    max_resolution[j],
                    fig_of_merit[j],
                    amp_contrast[0],
                    plot_path,
                )
            )
        return micrograph_list

    @staticmethod
    def for_cache(ctfmicrograph):
        return str(ctfmicrograph.micrograph_name)

    @staticmethod
    def db_unpack(micrograph_list):
        res = [
            {
                "micrograph_full_path": micrograph.micrograph_name,
                "astigmatism": micrograph.astigmatism,
                "astigmatism_angle": micrograph.defocus_angle,
                "estimated_resolution": micrograph.max_resolution,
                "estimated_defocus": (
                    float(micrograph.defocus_u) + float(micrograph.defocus_v)
                )
                / 2,
                "cc_value": micrograph.fig_of_merit,
                "amplitude_contrast": micrograph.amp_contrast,
                "fft_theoretical_full_path": micrograph.diagnostic_plot_path,
            }
            for micrograph in micrograph_list
        ]
        return res



from __future__ import annotations

import sys
from optparse import SUPPRESS_HELP, OptionParser
from pprint import pprint

from workflows.transport.stomp_transport import StompTransport

import relion


def run():
    parser = OptionParser()

    parser.add_option("-?", action="help", help=SUPPRESS_HELP)

    parser.add_option(
        "--test",
        action="store_true",
        dest="test",
        help="Run in ActiveMQ testing namespace (zocdev, default)",
    )
    parser.add_option(
        "--live",
        action="store_true",
        dest="test",
        help="Run in ActiveMQ live namespace (zocalo)",
    )

    parser.add_option(
        "--d",
        "--dir",
        "--directory",
        action="store",
        dest="relion_directory",
        help="Path to directory containing Relion data. Defaults to current directory",
        default=".",
    )
    # change settings when in live mode
    default_configuration = "/dls_sw/apps/zocalo/secrets/credentials-testing.cfg"
    if "--live" in sys.argv:
        default_configuration = "/dls_sw/apps/zocalo/secrets/credentials-live.cfg"

    StompTransport.load_configuration_file(default_configuration)
    StompTransport.add_command_line_options(parser)
    (options, args) = parser.parse_args(sys.argv[1:])
    stomp = StompTransport()
    stomp.connect()

    project = relion.Project(options.relion_directory)
    ctf_result = collect_ctffind(project)
    for item in ctf_result:
        message = {"parameters": item, "content": "dummy_content"}
        pprint(message)
        stomp.send("ispyb_connector", message)

    mc_result = collect_motion_correction(project)
    for item in mc_result:
        message = {"parameters": item, "content": "dummy_content"}
        pprint(message)
        stomp.send("ispyb_connector", message)

    class2d_result = collect_class2d(project)
    for item in class2d_result:
        message = {"parameters": item, "content": "dummy_content"}
        pprint(message)
        stomp.send("ispyb_connector", message)

    class3_result = collect_class3d(project)
    for item in class3_result:
        message = {"parameters": item, "content": "dummy_content"}
        pprint(message)
        stomp.send("ispyb_connector", message)


def collect_ctffind(project):
    ctf_dictionary_list = []
    for job in project.ctffind.values():
        for item in job:
            ctf_dictionary_list.append(
                {
                    "ispyb_command": "insert_ctf",
                    "astigmatism": item.astigmatism,
                    "astigmatism_angle": item.defocus_angle,
                    "max_estimated_resolution": item.max_resolution,
                    "estimated_defocus": (float(item.defocus_u) + float(item.defocus_v))
                    / 2,
                    "micrograph_name": item.micrograph_name,
                    "cc_value": item.fig_of_merit,
                }
            )
    return ctf_dictionary_list


def collect_motion_correction(project):
    motion_corr_dictionary_list = []
    for job in project.motioncorrection.values():
        for item in job:
            motion_corr_dictionary_list.append(
                {
                    "ispyb_command": "insert_motion_correction",
                    "micrograph_name": item.micrograph_name,
                    "total_motion": item.total_motion,
                    "early_motion": item.early_motion,
                    "late_motion": item.late_motion,
                    "average_motion_per_frame": (
                        float(item.total_motion)
                    ),  # / number of frames
                }
            )
    return motion_corr_dictionary_list


def collect_class2d(project):
    class2d_dictionary_list = []
    for job in project.class2D.values():
        for item in job:
            class2d_dictionary_list.append(
                {
                    "ispyb_command": "insert_class2d",
                    "reference_image": item.reference_image,
                }
            )
    return class2d_dictionary_list


def collect_class3d(project):
    class3d_dictionary_list = []
    for job in project.class3D.values():
        for item in job:
            class3d_dictionary_list.append(
                {
                    "ispyb_command": "insert_class3d",
                    "reference_image": item.reference_image,
                }
            )
    return class3d_dictionary_list


if __name__ == "__main__":
    run()



from __future__ import annotations

import logging
import pathlib
from typing import List

from relion._parser.autopick import ParticleCacheRecord, ParticlePickerInfo
from relion._parser.jobtype import JobType

logger = logging.getLogger("relion._parser.cryolo")


class Cryolo(JobType):
    def __init__(self, path, particle_cache=None, seen_star_files=None):
        super().__init__(path)
        self._particle_cache = particle_cache or {}

    def __eq__(self, other):
        if isinstance(other, Cryolo):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.Cryolo", self._basepath))

    def __repr__(self):
        return f"Cryolo({repr(str(self._basepath))})"

    def __str__(self):
        return f"<Cryolo parser at {self._basepath}>"

    @property
    def jobs(self):
        return sorted(
            d.resolve().name
            for d in self._basepath.iterdir()
            if d.is_symlink() and "crYOLO" in str(d)
        )

    def _load_job_directory(
        self, jobdir, input_tag: str = "in_mic", star_location: str = "Movies"
    ):
        num_particles = 0
        particles_per_micrograph = {}
        first_mic = ""
        coords = {}
        for star_file in (self._basepath / jobdir / star_location).glob("**/*"):
            if (
                star_file.is_file()
                and "gain" not in str(star_file)
                and "copy" not in str(star_file)
            ):
                part_coords = self._get_particle_info(jobdir, star_file)
                if part_coords:
                    coords[star_file] = part_coords
                    if particles_per_micrograph == {}:
                        first_mic = star_file
                    particles_per_micrograph[star_file] = len(part_coords)
                    num_particles += len(part_coords)

        # all of this just tracks back to a micrograph name from the MotionCorrection job
        try:
            jobfile = self._read_star_file(jobdir, "job.star")
        except (RuntimeError, FileNotFoundError, OSError, ValueError):
            return []
        info_table = self._find_table_from_column_name("_rlnJobOptionVariable", jobfile)
        if info_table is None:
            logger.debug(f"_rlnJobOptionVariable not found in file {jobfile}")
            return []
        variables = self.parse_star_file("_rlnJobOptionVariable", jobfile, info_table)
        inmicindex = variables.index(input_tag)
        ctffilename = pathlib.Path(
            self.parse_star_file("_rlnJobOptionValue", jobfile, info_table)[
                inmicindex
            ].replace("'", "")
        )
        ctffile = self._read_star_file_from_proj_dir(
            ctffilename.parts[0], ctffilename.relative_to(ctffilename.parts[0])
        )
        info_table = self._find_table_from_column_name("_rlnMicrographName", ctffile)
        if info_table is None:
            logger.debug(f"_rlnMicrographName not found in file {ctffile}")
            return []
        micrograph_names = self.parse_star_file(
            "_rlnMicrographName", ctffile, info_table
        )
        indexed_micrograph_names = {
            pathlib.Path(mn).stem: mn for mn in micrograph_names
        }

        particle_picker_info = []
        for mic, num_particles in particles_per_micrograph.items():
            mc_mic_name = self._get_micrograph_name(mic, indexed_micrograph_names)
            if not mc_mic_name:
                continue
            mic_parts = pathlib.Path(mc_mic_name).parts
            highlighted_micrograph = self._get_highlighted_mircograph(
                jobdir, mc_mic_name, mic_parts
            )
            particle_picker_info.append(
                ParticlePickerInfo(
                    num_particles,
                    mc_mic_name,
                    str(self._basepath.parent / mc_mic_name).replace(".mrc", ".jpeg"),
                    str(first_mic.relative_to(self._basepath / jobdir))
                    .replace("_autopick", "")
                    .replace(".star", ".mrc"),
                    str(highlighted_micrograph),
                    coords[mic],
                    jobdir,
                )
            )

        return particle_picker_info

    def _get_highlighted_mircograph(
        self, jobdir: str, mc_mic_name: str, mic_parts: List[str]
    ) -> pathlib.Path:
        return (
            self._basepath
            / jobdir
            / pathlib.Path(mc_mic_name)
            .relative_to(pathlib.Path(mic_parts[0]) / mic_parts[1])
            .with_suffix(".jpeg")
        )

    def _get_particle_info(self, jobdir, star_file):
        if self._particle_cache.get(jobdir) is None:
            self._particle_cache[jobdir] = {}
        if self._particle_cache[jobdir].get(star_file):
            try:
                if (
                    self._particle_cache[jobdir][star_file].file_size
                    == star_file.stat().st_size
                ):
                    return self._particle_cache[jobdir][star_file].data
            except FileNotFoundError:
                logger.debug(
                    "Could not find expected file containing particle data",
                    exc_info=True,
                )
                return []

        try:
            file = self._read_star_file(
                jobdir, star_file.relative_to(self._basepath / jobdir)
            )
        except (RuntimeError, FileNotFoundError, OSError, ValueError):
            return []

        info_table = self._find_table_from_column_name("_rlnCoordinateX", file)
        if info_table is None:
            logger.debug(f"_rlnCoordinateX not found in file {file}")
            return []

        all_particles = self.parse_star_file("_rlnCoordinateX", file, info_table)
        ys = self.parse_star_file("_rlnCoordinateY", file, info_table)
        coords = [(x, y) for x, y in zip(all_particles, ys)]
        try:
            self._particle_cache[jobdir][star_file] = ParticleCacheRecord(
                coords,
                star_file.stat().st_size,
            )
        except FileNotFoundError:
            return []
        return coords

    def _get_micrograph_name(self, micrograph, micrograph_names):
        return micrograph_names.get(micrograph.stem.replace("_autopick", ""))

    @staticmethod
    def for_cache(partpickinfo):
        return str(partpickinfo.number_of_particles)

    @staticmethod
    def db_unpack(partpickinfo):
        res = [
            {
                "number_of_particles": pi.number_of_particles,
                "job_string": pi.job,
                "micrograph_full_path": pi.micrograph_full_path,
                "mc_image_full_path": pi.mc_image_full_path,
                "summary_image_full_path": pi.highlighted_micrograph,
                "particle_coordinates": pi.coordinates,
            }
            for pi in partpickinfo
        ]
        return res


class CryoloAutoPick(Cryolo):
    def __eq__(self, other):
        if isinstance(other, CryoloAutoPick):  # check this
            return self._basepath == other._basepath
        return False

    def __hash__(self):
        return hash(("relion._parser.CryoloAutoPick", self._basepath))

    def __repr__(self):
        return f"CryoloAutoPick({repr(str(self._basepath))})"

    def __str__(self):
        return f"<CryoloAutoPick parser at {self._basepath}>"

    def _load_job_directory(self, jobdir):
        return super()._load_job_directory(
            jobdir, input_tag="'input_file'", star_location="STAR"
        )

    def _get_highlighted_mircograph(
        self, jobdir: str, mc_mic_name: str, mic_parts: List[str]
    ) -> pathlib.Path:
        return (
            self._basepath / jobdir / "STAR" / pathlib.Path(mc_mic_name).name
        ).with_suffix(".jpeg")

    @property
    def jobs(self):
        return super(Cryolo, self).jobs



from __future__ import annotations

import argparse
from pathlib import Path

import procrunner


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p",
        "--project",
        help="Path to original RELION project directory",
        dest="project",
    )
    parser.add_argument(
        "-d",
        "--destination",
        help="Path to directory where new project should be created",
        dest="destination",
        default=".",
    )
    parser.add_argument(
        "--hardlink-cutoff",
        type=int,
        help="Size in bytes past which a hardlink will be made to a file rather than copying",
        default=5e6,
    )
    args = parser.parse_args()

    required_dirs = (".Nodes", ".TMP_runfiles")

    ignore_file_extensions = {"Class2D": [".jpeg"]}

    project_path = Path(args.project).resolve()
    destination_path = Path(args.destination).resolve()

    if not destination_path.exists():
        destination_path.mkdir(exist_ok=True)

    for f in project_path.glob("*"):
        if (f.is_dir() and f.name in required_dirs) or f.is_file():
            procrunner.run(
                [
                    "rsync",
                    "--recursive",
                    str(f),
                    str(destination_path)
                    if f.is_dir()
                    else str(destination_path / f.relative_to(project_path)),
                ]
            )
        elif f.is_dir():
            job_type_dir = destination_path / f.relative_to(project_path)
            job_type_dir.mkdir(exist_ok=True)
            for job_dir in f.glob("*"):
                if not job_dir.is_symlink():
                    new_job_dir = destination_path / job_dir.relative_to(project_path)
                    new_job_dir.mkdir(exist_ok=True)
                    for jf in job_dir.glob("*"):
                        if jf.is_file():
                            if jf.suffix in ignore_file_extensions.get(f.name, []):
                                continue
                            file_linked = False
                            if jf.stat().st_size > args.hardlink_cutoff:
                                try:
                                    jf.link_to(
                                        f"{destination_path / jf.relative_to(project_path)}"
                                    )
                                    file_linked = True
                                except OSError:
                                    file_linked = False
                            if not file_linked:
                                procrunner.run(
                                    [
                                        "rsync",
                                        str(jf),
                                        f"{destination_path / jf.relative_to(project_path)}",
                                    ]
                                )
                        else:
                            try:
                                (
                                    destination_path / jf.relative_to(project_path)
                                ).symlink_to(jf)
                            except FileExistsError:
                                pass
        elif f.is_symlink():
            source = f.resolve()
            try:
                (destination_path / f.relative_to(project_path)).symlink_to(source)
            except FileExistsError:
                pass



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict

from relion.cryolo_relion_it.cryolo_relion_it import RelionItOptions


def generate_pipeline_options(
    relion_it_options: RelionItOptions, submission_types: Dict[str, str]
) -> dict:
    for queue in submission_types.values():
        if queue not in ("gpu", "cpu", "gpu-smp", "cpu-smp", ""):
            raise ValueError(
                f'The queue for a job must be either gpu, cpu, gpu-smp, cpu-smp or "", not {queue}'
            )

    queue_options_gpu = {
        "do_queue": "Yes",
        "qsubscript": relion_it_options.queue_submission_template,
    }
    queue_options_gpu_smp = {
        "do_queue": "Yes",
        "qsubscript": relion_it_options.queue_submission_template_smp,
    }
    queue_options_cpu = {
        "do_queue": "Yes",
        "qsubscript": relion_it_options.queue_submission_template_cpu,
    }
    queue_options_cpu_smp = {
        "do_queue": "Yes",
        "qsubscript": relion_it_options.queue_submission_template_cpu_smp,
    }
    queue_options: Dict[str, dict] = {
        "gpu": queue_options_gpu,
        "cpu": queue_options_cpu,
        "gpu-smp": queue_options_gpu_smp,
        "cpu-smp": queue_options_cpu_smp,
        "": {},
    }

    job_options: Dict[str, Any] = {}

    job_options["relion.import.movies"] = {
        "fn_in_raw": relion_it_options.import_images,
        "angpix": relion_it_options.angpix,
        "kV": relion_it_options.voltage,
    }

    job_options["relion.motioncorr.own"] = {
        "fn_defect": relion_it_options.motioncor_defectfile,
        "dose_per_frame": relion_it_options.motioncor_doseperframe,
        "fn_gain_ref": relion_it_options.motioncor_gainreference
        if Path(relion_it_options.motioncor_gainreference).exists()
        else "",
        "eer_grouping": relion_it_options.eer_grouping,
        "patch_x": relion_it_options.motioncor_patches_x,
        "patch_y": relion_it_options.motioncor_patches_y,
        "bfactor": relion_it_options.motioncor_bfactor,
        "bin_factor": relion_it_options.motioncor_binning,
        "gain_flip": relion_it_options.motioncor_gainflip,
        "gain_rot": relion_it_options.motioncor_gainrot,
        "other_args": f"{relion_it_options.motioncor_other_args} --skip_logfile",
        "nr_mpi": relion_it_options.motioncor_mpi,
        "nr_threads": relion_it_options.motioncor_threads,
        "do_float16": relion_it_options.motioncor_do_float16,
    }

    job_options["relion.motioncorr.motioncor2"] = {
        **job_options["relion.motioncorr.own"],
        "fn_motioncor2_exe": relion_it_options.motioncor_exe,
        "gpu_ids": "0:1:2:3",
        "other_motioncor2_args": relion_it_options.motioncor2_other_args,
    }

    job_options["icebreaker.micrograph_analysis.micrographs"] = {
        "nr_threads": relion_it_options.icebreaker_threads_number,
    }

    job_options["icebreaker.micrograph_analysis.enhancecontrast"] = {
        "nr_threads": relion_it_options.icebreaker_threads_number,
    }

    job_options["icebreaker.micrograph_analysis.summary"] = {
        "nr_threads": relion_it_options.icebreaker_threads_number,
    }

    job_options["relion.ctffind.ctffind4"] = {
        "dast": relion_it_options.ctffind_astigmatism,
        "box": relion_it_options.ctffind_boxsize,
        "dfmax": relion_it_options.ctffind_defocus_max,
        "dfmin": relion_it_options.ctffind_defocus_min,
        "dfstep": relion_it_options.ctffind_defocus_step,
        "resmax": relion_it_options.ctffind_maxres,
        "resmin": relion_it_options.ctffind_minres,
        "do_phaseshift": relion_it_options.ctffind_do_phaseshift,
        "nr_mpi": relion_it_options.ctffind_mpi,
    }

    job_options["relion.autopick.log"] = {
        "log_diam_min": relion_it_options.autopick_LoG_diam_min,
        "log_diam_max": relion_it_options.autopick_LoG_diam_max,
        "log_maxres": relion_it_options.autopick_lowpass,
        "log_adjust_thr": relion_it_options.autopick_LoG_adjust_threshold,
        "log_upper_thr": relion_it_options.autopick_LoG_upper_threshold,
        "gpu_ids": "0:1:2:3",
        "nr_mpi": relion_it_options.autopick_mpi,
    }

    job_options["relion.autopick.ref3d"] = {
        "ref3d_symmetry": relion_it_options.autopick_3dref_symmetry,
        "ref3d_sampling": relion_it_options.autopick_3dref_sampling,
        "lowpass": relion_it_options.autopick_lowpass,
        "angpix_ref": relion_it_options.autopick_ref_angpix,
        "threshold_autopick": relion_it_options.autopick_refs_threshold,
        "mindist_autopick": relion_it_options.autopick_refs_min_distance,
        "maxstddevnoise_autopick": relion_it_options.autopick_stddev_noise,
        "minavgnoise_autopick": relion_it_options.autopick_avg_noise,
        "gpu_ids": "0:1:2:3",
        "nr_mpi": relion_it_options.autopick_mpi,
    }

    job_options["cryolo.autopick"] = {
        "model_path": relion_it_options.cryolo_gmodel,
        "config_file": relion_it_options.cryolo_config,
        "box_size": int(
            relion_it_options.extract_boxsize / relion_it_options.motioncor_binning
        ),
        "confidence_threshold": relion_it_options.cryolo_threshold,
        "gpus": relion_it_options.cryolo_pick_gpus,
    }

    bin_corrected_box_size = int(
        relion_it_options.extract_boxsize / relion_it_options.motioncor_binning
    )
    job_options["relion.extract"] = {
        "bg_diameter": relion_it_options.extract_bg_diameter,
        "extract_size": bin_corrected_box_size + bin_corrected_box_size % 2,
        "do_rescale": bool(relion_it_options.extract_downscale),
        "rescale": relion_it_options.extract_small_boxsize,
        "nr_mpi": relion_it_options.extract_mpi,
    }

    job_options["relion.select.split"] = {
        "split_size": relion_it_options.batch_size,
    }

    job_options["icebreaker.micrograph_analysis.particles"] = {
        "nr_threads": relion_it_options.icebreaker_threads_number,
    }

    job_options["relion.class2d.em"] = {
        "nr_classes": relion_it_options.class2d_nr_classes,
        "nr_iter_em": relion_it_options.class2d_nr_iter,
        "psi_sampling": relion_it_options.class2d_angle_step,
        "offset_range": relion_it_options.class2d_offset_range,
        "offset_step": relion_it_options.class2d_offset_step,
        "ctf_intact_first_peak": relion_it_options.class2d_ctf_ign1stpeak,
        "do_preread_images": relion_it_options.refine_preread_images,
        "particle_diameter": relion_it_options.mask_diameter,
        "scratch_dir": relion_it_options.refine_scratch_disk,
        "nr_pool": relion_it_options.refine_nr_pool,
        "use_gpu": relion_it_options.refine_do_gpu,
        "gpu_ids": relion_it_options.refine_gpu,
        "nr_mpi": relion_it_options.refine_mpi,
        "nr_threads": relion_it_options.refine_threads,
    }

    job_options["relion.class2d.vdam"] = {
        "nr_classes": relion_it_options.class2d_nr_classes,
        "nr_iter_grad": relion_it_options.class2d_nr_iter,
        "psi_sampling": relion_it_options.class2d_angle_step,
        "offset_range": relion_it_options.class2d_offset_range,
        "offset_step": relion_it_options.class2d_offset_step,
        "ctf_intact_first_peak": relion_it_options.class2d_ctf_ign1stpeak,
        "do_preread_images": relion_it_options.refine_preread_images,
        "particle_diameter": relion_it_options.mask_diameter,
        "scratch_dir": relion_it_options.refine_scratch_disk,
        "nr_pool": relion_it_options.refine_nr_pool,
        "use_gpu": relion_it_options.refine_do_gpu,
        "gpu_ids": "0,1,2,3",
        "nr_mpi": 1,
        "nr_threads": relion_it_options.inimodel_threads,
    }

    job_options["relion.initialmodel"] = {
        "nr_classes": relion_it_options.inimodel_nr_classes,
        "sym_name": relion_it_options.symmetry,
        "do_run_C1": relion_it_options.inimodel_start_c1,
        "sampling": relion_it_options.inimodel_angle_step,
        "offset_step": relion_it_options.inimodel_offset_step,
        "offset_range": relion_it_options.inimodel_offset_range,
        "nr_iter": relion_it_options.inimodel_nr_iter_inbetween,
        "particle_diameter": relion_it_options.mask_diameter,
        "do_preread_images": relion_it_options.refine_preread_images,
        "scratch_dir": relion_it_options.refine_scratch_disk,
        "nr_pool": relion_it_options.refine_nr_pool,
        "use_gpu": relion_it_options.refine_do_gpu,
        "gpu_ids": "0,1,2,3",
        "nr_threads": relion_it_options.inimodel_threads,
    }

    job_options["relion.class3d"] = {
        "fn_mask": relion_it_options.class3d_reference_mask,
        "nr_classes": relion_it_options.class3d_nr_classes,
        "sym_name": relion_it_options.symmetry,
        "ini_high": relion_it_options.class3d_ini_lowpass,
        "tau_fudge": relion_it_options.class3d_T_value,
        "particle_diameter": relion_it_options.mask_diameter,
        "nr_iter": relion_it_options.class3d_nr_iter,
        "sampling": relion_it_options.class3d_angle_step,
        "offset_range": relion_it_options.class3d_offset_range,
        "offset_step": relion_it_options.class3d_offset_step,
        "ref_correct_greyscale": relion_it_options.class3d_ref_is_correct_greyscale,
        "do_ctf_correction": relion_it_options.class3d_ref_is_ctf_corrected,
        "ctf_intact_first_peak": relion_it_options.class3d_ctf_ign1stpeak,
        "do_preread_images": relion_it_options.refine_preread_images,
        "use_gpu": relion_it_options.refine_do_gpu,
        "gpu_ids": relion_it_options.refine_gpu,
        "nr_mpi": relion_it_options.refine_mpi,
        "nr_threads": relion_it_options.refine_threads,
    }

    pipeline_options = {
        key: {**job_options[key], **queue_options[q]}
        for key, q in submission_types.items()
    }
    return pipeline_options
